/**
 ******************************************************************************
 * Copyright (c), Takenori Imoto
 *  software http://www.kaede-software.com/
 * All rights reserved.
 ******************************************************************************
 * \[XR[h`oCi`AύX邩Ȃ킸Aȉ̏
 * ꍇɌAĔЕzюgp܂B
 *
 * E\[XR[hĔЕzꍇAL̒쌠\A{ꗗAщLƐ
 *   ܂߂邱ƁB
 * EoCi`ōĔЕzꍇAЕzɕt̃hLg̎ɁAL
 *   쌠\A{ꗗAщLƐӏ܂߂邱ƁB
 * Eʂɂʂ̋ȂɁA{\tgEFAhi̐`܂͔̔
 *   iɁAgD̖O܂̓Rgr[^[̖OgpĂ͂ȂȂB
 *
 * {\tgEFÁA쌠҂уRgr[^[ɂāû܂܁v
 * ĂAَ킸AƓIȎgp\Aѓ̖ړIɑ΂K
 * ɊւÖق̕ۏ؂܂߁A܂Ɍ肳ȂAȂۏ؂܂
 * B쌠҂Rgr[^[AR̂킸AQ̌
 * 킸AӔC̍_ł邩iӔCł邩iߎ̑́js@
 * sׂł邩킸Aɂ̂悤ȑQ\m炳ĂƂ
 * ĂA{\tgEFA̎gpɂĔi֕i܂͑pT[rX̒BA
 * gp̑rAf[^̑rAv̑rAƖ̒f܂߁A܂Ɍ肳
 * jڑQAԐڑQAIȑQAʑQAIQA܂͌ʑQ
 * āAؐӔC𕉂Ȃ̂Ƃ܂B
 ******************************************************************************
 * {\tgEFÁAgg2 ( http://kikyou.info/tvp/ ) ̃\[XR[hJava
 * ɏ̂ꕔgpĂ܂B
 * gg2 Copyright (C) W.Dee <dee@kikyou.info> and contributors
 ******************************************************************************
 */
package jp.kirikiri.tvp2.base;

import java.lang.ref.WeakReference;

import jp.kirikiri.tjs2.Error;
import jp.kirikiri.tjs2.Dispatch2;
import jp.kirikiri.tjs2.NativeInstanceObject;
import jp.kirikiri.tjs2.TJS;
import jp.kirikiri.tjs2.TJSException;
import jp.kirikiri.tjs2.Variant;
import jp.kirikiri.tjs2.VariantClosure;
import jp.kirikiri.tjs2.VariantException;
import jp.kirikiri.tvp2.TVP;

/**
 * AsyncTrigger is a class for invoking events at asynchronous.
 * Script can trigger event but the event is not delivered immediately,
 * is delivered at next event flush phase.
 *
 * TJS AsyncTrigger native instance
 */
public class AsyncTriggerNI extends NativeInstanceObject {
	private static final int atmNormal = 0, atmExclusive = 1, atmAtIdle = 2;

	static private final String ON_FIRE = "onFire";

	private WeakReference<Dispatch2> mOwner;
	private VariantClosure mActionOwner; // object to send action
	private String mActionName;

	private boolean mCached; // cached operation
	private int mMode; // event mode
	private int mIdlePendingCount;

	public AsyncTriggerNI() {
		//mOwner = null;
		mCached = true;
		//mIdlePendingCount = 0;
		mMode = atmNormal;
		mActionOwner = new VariantClosure(null);
		//mActionOwner.mObject = mActionOwner.mObjThis = null;
		mActionName = EventManager.ActionName;
	}
	public int construct( Variant[] param, Dispatch2 tjs_obj) throws TJSException {
		if( param.length < 1) return Error.E_BADPARAMCOUNT;

		int hr = super.construct( param, tjs_obj);
		if( hr < 0 ) return hr;

		if( param.length >= 2 && param[1].isVoid() != true )
			mActionName = param[1].asString(); // action function to be called

		mActionOwner.set( param[0].asObjectClosure() );
		mOwner = new WeakReference<Dispatch2>(tjs_obj);

		return Error.S_OK;
	}
	public void invalidate() throws VariantException, TJSException {
		TVP.EventManager.cancelSourceEvents(mOwner.get());
		mOwner.clear();

		mActionOwner.mObjThis = mActionOwner.mObject = null;
		//mActionOwner = null;

		super.invalidate();
	}

	public VariantClosure getActionOwner() { return mActionOwner; }
	public String getActionName() { return mActionName; }

	public void trigger() {
		// trigger event
		Dispatch2 owner = mOwner.get();
		if(owner!=null) {
			if(mCached) {
				// remove undelivered events from queue when "Cached" flag is set
				TVP.EventManager.cancelSourceEvents(owner);
			}

			int flags = EventManager.EPT_POST;
			if(mMode == atmExclusive) flags |= EventManager.EPT_EXCLUSIVE;  // fire exclusive event
			if(mMode == atmAtIdle)    flags |= EventManager.EPT_IDLE;       // fire idle event

			TVP.EventManager.postEvent( owner, owner, ON_FIRE, 0, flags, TJS.NULL_ARG );
		}
	}
	public 	void cancel() {
		// cancel event
		Dispatch2 owner = mOwner.get();
		if(owner!=null) TVP.EventManager.cancelSourceEvents(owner);
		mIdlePendingCount = 0;
	}

	public boolean getCached() { return mCached; }
	public void setCached(boolean b) {
		// set cached operation flag.
		// when this flag is set, only one event is delivered at once.
		if(mCached != b) {
			mCached = b;
			cancel(); // all events are canceled
		}
	}

	public int getMode() { return mMode; }
	public void setMode( int m) {
		if( mMode != m) {
			mMode = m;
			cancel();
		}
	}
}
